import { parser } from '@tk-tool/lezer-spel';
import { syntaxTree, LRLanguage, indentNodeProp, continuedIndent, flatIndent, delimitedIndent, foldNodeProp, foldInside, LanguageSupport } from '@codemirror/language';
import { snippetCompletion, ifNotIn, completeFromList } from '@codemirror/autocomplete';
import { NodeWeakMap, IterMode } from '@tk-tool/lezer-common';

/**
A collection of JavaScript-related
[snippets](https://codemirror.net/6/docs/ref/#autocomplete.snippet).
*/
// 片段定义
const snippets = [
    /*@__PURE__*/snippetCompletion("#stringJoinLoop()", {
        label: "#stringJoinLoop",
        detail: "#stringJoinLoop()",
        type: "keyword"
    }),
    // snip("function ${name}(${params}) {\n\t${}\n}", {
    //   label: "function",
    //   detail: "definition",
    //   type: "keyword"
    // }),
    /*@__PURE__*/snippetCompletion("for (let ${index} = 0; ${index} < ${bound}; ${index}++) {\n\t${}\n}", {
        label: "for",
        detail: "loop",
        type: "keyword"
    }),
    /*@__PURE__*/snippetCompletion("for (let ${name} of ${collection}) {\n\t${}\n}", {
        label: "for",
        detail: "of loop",
        type: "keyword"
    }),
    /*@__PURE__*/snippetCompletion("do {\n\t${}\n} while (${})", {
        label: "do",
        detail: "loop",
        type: "keyword"
    }),
    /*@__PURE__*/snippetCompletion("while (${}) {\n\t${}\n}", {
        label: "while",
        detail: "loop",
        type: "keyword"
    }),
    /*@__PURE__*/snippetCompletion("try {\n\t${}\n} catch (${error}) {\n\t${}\n}", {
        label: "try",
        detail: "/ catch block",
        type: "keyword"
    }),
    /*@__PURE__*/snippetCompletion("if (${}) {\n\t${}\n}", {
        label: "if",
        detail: "block",
        type: "keyword"
    }),
    /*@__PURE__*/snippetCompletion("if (${}) {\n\t${}\n} else {\n\t${}\n}", {
        label: "if",
        detail: "/ else block",
        type: "keyword"
    }),
    // snip("class ${name} {\n\tconstructor(${params}) {\n\t\t${}\n\t}\n}", {
    //   label: "class",
    //   detail: "definition",
    //   type: "keyword"
    // }),
    // snip("import {${names}} from \"${module}\"\n${}", {
    //   label: "import",
    //   detail: "named",
    //   type: "keyword"
    // }),
    // snip("import ${name} from \"${module}\"\n${}", {
    //   label: "import",
    //   detail: "default",
    //   type: "keyword"
    // })
];

const cache = /*@__PURE__*/new NodeWeakMap();
const ScopeNodes = /*@__PURE__*/new Set([
    "Script", "Block",
    "FunctionExpression", "FunctionDeclaration", "ArrowFunction", "MethodDeclaration",
    "ForStatement"
]);
function defID(type) {
    return (node, def) => {
        const id = node.node.getChild("VariableDefinition");
        if (id)
            def(id, type);
        return true;
    };
}
const functionContext = ["FunctionDeclaration"];
const gatherCompletions = {
    FunctionDeclaration: /*@__PURE__*/defID("function"),
    ClassDeclaration: /*@__PURE__*/defID("class"),
    ClassExpression: () => true,
    EnumDeclaration: /*@__PURE__*/defID("constant"),
    TypeAliasDeclaration: /*@__PURE__*/defID("type"),
    NamespaceDeclaration: /*@__PURE__*/defID("namespace"),
    VariableDefinition(node, def) { if (!node.matchContext(functionContext))
        def(node, "variable"); },
    TypeDefinition(node, def) { def(node, "type"); },
    __proto__: null
};
function getScope(doc, node) {
    const cached = cache.get(node);
    if (cached)
        return cached;
    const completions = [];
    let top = true;
    function def(node, type) {
        const name = doc.sliceString(node.from, node.to);
        completions.push({ label: name, type });
    }
    node.cursor(IterMode.IncludeAnonymous).iterate(node => {
        if (top) {
            top = false;
        }
        else if (node.name) {
            const gather = gatherCompletions[node.name];
            if (gather && gather(node, def) || ScopeNodes.has(node.name))
                return false;
        }
        else if (node.to - node.from > 8192) {
            // Allow caching for bigger internal nodes
            for (const c of getScope(doc, node.node))
                completions.push(c);
            return false;
        }
    });
    cache.set(node, completions);
    return completions;
}
const Identifier = /^[\w$\xa1-\uffff][\w$\d\xa1-\uffff]*$/;
const dontComplete = [
    "TemplateString", "String", "RegExp",
    "LineComment", "BlockComment",
    "VariableDefinition", "TypeDefinition", "Label",
    "PropertyDefinition", "PropertyName",
    "PrivatePropertyDefinition", "PrivatePropertyName"
];
/**
Completion source that looks up locally defined names in 中查找本地定义名称的完成源
JavaScript code.
*/
/*@__PURE__*/console.log({
    "inner init": 'inner init'
});
// 调用时机是 每次录入数据时
function localCompletionSource(context) {
    const inner = syntaxTree(context.state).resolveInner(context.pos, -1);
    console.log({
        inner
    });
    if (dontComplete.indexOf(inner.name) > -1)
        return null;
    const isWord = inner.name == "VariableName" || inner.to - inner.from < 20 && Identifier.test(context.state.sliceDoc(inner.from, inner.to));
    if (!isWord && !context.explicit)
        return null;
    let options = [];
    for (let pos = inner; pos; pos = pos.parent) {
        if (ScopeNodes.has(pos.name))
            options = options.concat(getScope(context.state.doc, pos));
    }
    return {
        options,
        from: isWord ? inner.from : context.pos,
        validFor: Identifier
    };
}
function pathFor(read, member, name) {
    var _a;
    const path = [];
    for (;;) {
        const obj = member.firstChild;
        let prop;
        if ((obj === null || obj === void 0 ? void 0 : obj.name) == "VariableName") {
            path.push(read(obj));
            return { path: path.reverse(), name };
        }
        else if ((obj === null || obj === void 0 ? void 0 : obj.name) == "MemberExpression" && ((_a = (prop = obj.lastChild)) === null || _a === void 0 ? void 0 : _a.name) == "PropertyName") {
            path.push(read(prop));
            member = obj;
        }
        else {
            return null;
        }
    }
}
/**
Helper function for defining JavaScript completion sources. It
returns the completable name and object path for a completion
context, or null if no name/property completion should happen at
that position. For example, when completing after `a.b.c` it will
return `{path: ["a", "b"], name: "c"}`. When completing after `x`
it will return `{path: [], name: "x"}`. When not in a property or
name, it will return null if `context.explicit` is false, and
`{path: [], name: ""}` otherwise.
*/
function completionPath(context) {
    const read = (node) => context.state.doc.sliceString(node.from, node.to);
    const inner = syntaxTree(context.state).resolveInner(context.pos, -1);
    if (inner.name == "PropertyName") {
        return pathFor(read, inner.parent, read(inner));
    }
    else if (dontComplete.indexOf(inner.name) > -1) {
        return null;
    }
    else if (inner.name == "VariableName" || inner.to - inner.from < 20 && Identifier.test(read(inner))) {
        return { path: [], name: read(inner) };
    }
    else if ((inner.name == "." || inner.name == "?.") && inner.parent.name == "MemberExpression") {
        return pathFor(read, inner.parent, "");
    }
    else if (inner.name == "MemberExpression") {
        return pathFor(read, inner, "");
    }
    else {
        return context.explicit ? { path: [], name: "" } : null;
    }
}
function enumeratePropertyCompletions(obj, top) {
    const options = [];
    const seen = new Set();
    for (let depth = 0;; depth++) {
        for (const name of (Object.getOwnPropertyNames || Object.keys)(obj)) {
            if (seen.has(name))
                continue;
            seen.add(name);
            let value;
            try {
                value = obj[name];
            }
            catch (_) {
                continue;
            }
            options.push({
                label: name,
                type: typeof value == "function"
                    ? (/^[A-Z]/.test(name) ? "class" : top ? "function" : "method")
                    : top ? "variable" : "property",
                boost: -depth
            });
        }
        const next = Object.getPrototypeOf(obj);
        if (!next)
            return options;
        obj = next;
    }
}
/**
Defines a [completion source](https://codemirror.net/6/docs/ref/#autocomplete.CompletionSource) that
completes from the given scope object (for example `globalThis`).
Will enter properties of the object when completing properties on
a directly-named path.
*/
function scopeCompletionSource(scope) {
    const cache = new Map();
    return (context) => {
        const path = completionPath(context);
        if (!path)
            return null;
        let target = scope;
        for (const step of path.path) {
            target = target[step];
            if (!target)
                return null;
        }
        let options = cache.get(target);
        if (!options)
            cache.set(target, options = enumeratePropertyCompletions(target, !path.path.length));
        return {
            from: context.pos - path.name.length,
            options,
            validFor: Identifier
        };
    };
}

/**
A language provider based on the [Lezer JavaScript
parser](https://github.com/lezer-parser/spel), extended with
highlighting and indentation information.
*/
// TODO spelLanguage中 存在 parser的初始化
// 次返回值是一个 Language 类型数据
const spelLanguage = /*@__PURE__*/LRLanguage.define({
    name: "spel",
    parser: /*@__PURE__*/parser.configure({
        props: [
            // indentNodeProp 将计算缩进的函数与节点类型关联起来。 向函数传递一个上下文对象，其中包含相关值和一些与缩进相关的帮助器方法。
            /*@__PURE__*/indentNodeProp.add({
                IfStatement: /*@__PURE__*/continuedIndent({ except: /^\s*({|else\b)/ }),
                TryStatement: /*@__PURE__*/continuedIndent({ except: /^\s*({|catch\b|finally\b)/ }),
                LabeledStatement: flatIndent,
                SwitchBody: context => {
                    const after = context.textAfter;
                    const closed = /^\s*\}/.test(after);
                    const isCase = /^\s*(case|default)\b/.test(after);
                    return context.baseIndent + (closed ? 0 : isCase ? 1 : 2) * context.unit;
                },
                Block: /*@__PURE__*/delimitedIndent({ closing: "}" }),
                ArrowFunction: cx => cx.baseIndent + cx.unit,
                "TemplateString BlockComment": () => null,
                "Statement Property": /*@__PURE__*/continuedIndent({ except: /^{/ })
            }),
            // 折叠相关 最后，foldNodeProp将折叠信息与节点类型关联起来。我们允许通过隐藏除分隔符以外的所有内容来折叠应用程序节点。
            /*@__PURE__*/foldNodeProp.add({
                "Block ClassBody SwitchBody EnumBody ObjectExpression ArrayExpression": foldInside,
                BlockComment(tree) { return { from: tree.from + 2, to: tree.to - 2 }; }
            })
        ]
    }),
    languageData: {
        closeBrackets: { brackets: ["(", "[", "{", "'", '"', "`"] },
        // commentTokens: { line: "//", block: { open: "/*", close: "*/" } },
        indentOnInput: /^\s*(?:case |default:|\{|\}|<\/)$/,
        wordChars: "$"
    }
});
/**
A language provider for TypeScript.
*/
// export const typescriptLanguage = spelLanguage.configure({ dialect: "ts" }, "typescript")
/**
Language provider for JSX.
*/
// export const jsxLanguage = spelLanguage.configure({ dialect: "jsx" })
/**
Language provider for JSX + TypeScript.
*/
// export const tsxLanguage = spelLanguage.configure({ dialect: "jsx ts" }, "typescript")
const keywords = /*@__PURE__*/"break case const continue default delete export extends false finally in instanceof let new return static super switch this throw true typeof var yield tk tk1 tk2 tk3 #stringJoinLoop #filterLoop".split(" ").map(kw => ({ label: kw, type: "keyword" }));
/**
JavaScript support. Includes [snippet](https://codemirror.net/6/docs/ref/#lang-javascript.snippets)
completion.
*/
function spel() {
    // const lang = spelLanguage
    // 父级存在 startParse 函数
    console.log({
        spelLanguage
    });
    // 传入的是一个 Facet 类型的入参
    return new LanguageSupport(spelLanguage, [
        // 其中调用 FacetProvider 方法
        // 次返回值是一个 Facet 类型数据
        spelLanguage.data.of({
            // 支持扩充提示模板
            autocomplete: ifNotIn(dontComplete, completeFromList(snippets.concat(keywords)))
        }),
        spelLanguage.data.of({
            autocomplete: localCompletionSource
        }),
        // config.jsx ?    : []
        []
    ]);
}
// function findOpenTag (node: SyntaxNode) {
//   for (;;) {
//     if (node.name == "JSXOpenTag" || node.name == "JSXSelfClosingTag" || node.name == "JSXFragmentTag") return node
//     if (!node.parent) return null
//     node = node.parent
//   }
// }
// function elementName (doc: Text, tree: SyntaxNode | null | undefined, max = doc.length) {
//   for (let ch = tree?.firstChild; ch; ch = ch.nextSibling) {
//     if (ch.name == "JSXIdentifier" || ch.name == "JSXBuiltin" || ch.name == "JSXNamespacedName" ||
//         ch.name == "JSXMemberExpression") { return doc.sliceString(ch.from, Math.min(ch.to, max)) }
//   }
//   return ""
// }
// const android = typeof navigator == "object" && /Android\b/.test(navigator.userAgent)
/**
Extension that will automatically insert JSX close tags when a `>` or
`/` is typed.
*/
// export const    = EditorView.inputHandler.of((view, from, to, text) => {
//   if ((android ? view.composing : view.compositionStarted) || view.state.readOnly ||
//       from != to || (text != ">" && text != "/") ||
//       !spelLanguage.isActiveAt(view.state, from, -1)) return false
//   const { state } = view
//   const changes = state.changeByRange(range => {
//     const { head } = range; let around = syntaxTree(state).resolveInner(head, -1); let name
//     if (around.name == "JSXStartTag") around = around.parent!
//     if (text == ">" && around.name == "JSXFragmentTag") {
//       return { range: EditorSelection.cursor(head + 1), changes: { from: head, insert: `></>` } }
//     } else if (text == "/" && around.name == "JSXFragmentTag") {
//       const empty = around.parent; const base = empty?.parent
//       if (empty!.from == head - 1 && base!.lastChild?.name != "JSXEndTag" &&
//           (name = elementName(state.doc, base?.firstChild, head))) {
//         const insert = `/${name}>`
//         return { range: EditorSelection.cursor(head + insert.length), changes: { from: head, insert } }
//       }
//     } else if (text == ">") {
//       const openTag = findOpenTag(around)
//       if (openTag && openTag.lastChild?.name != "JSXEndTag" &&
//           state.sliceDoc(head, head + 2) != "</" &&
//           (name = elementName(state.doc, openTag, head))) { return { range: EditorSelection.cursor(head + 1), changes: { from: head, insert: `></${name}>` } } }
//     }
//     return { range }
//   })
//   if (changes.changes.empty) return false
//   view.dispatch(changes, { userEvent: "input.type", scrollIntoView: true })
//   return true
// })

/**
Connects an [ESLint](https://eslint.org/) linter to CodeMirror's
[lint](https://codemirror.net/6/docs/ref/#lint) integration. `eslint` should be an instance of the
[`Linter`](https://eslint.org/docs/developer-guide/nodejs-api#linter)
class, and `config` an optional ESLint configuration. The return
value of this function can be passed to [`linter`](https://codemirror.net/6/docs/ref/#lint.linter)
to create a JavaScript linting extension.

Note that ESLint targets node, and is tricky to run in the
browser. The
[eslint-linter-browserify](https://github.com/UziTech/eslint-linter-browserify)
package may help with that (see
[example](https://github.com/UziTech/eslint-linter-browserify/blob/master/example/script.js)).
*/
function esLint(eslint, config) {
    if (!config) {
        config = {
            parserOptions: { ecmaVersion: 2019, sourceType: "module" },
            env: { browser: true, node: true, es6: true, es2015: true, es2017: true, es2020: true },
            rules: {}
        };
        eslint.getRules().forEach((desc, name) => {
            if (desc.meta.docs.recommended)
                config.rules[name] = 2;
        });
    }
    return (view) => {
        const { state } = view;
        const found = [];
        for (const { from, to } of spelLanguage.findRegions(state)) {
            const fromLine = state.doc.lineAt(from);
            const offset = { line: fromLine.number - 1, col: from - fromLine.from, pos: from };
            for (const d of eslint.verify(state.sliceDoc(from, to), config)) {
                found.push(translateDiagnostic(d, state.doc, offset));
            }
        }
        return found;
    };
}
function mapPos(line, col, doc, offset) {
    return doc.line(line + offset.line).from + col + (line == 1 ? offset.col - 1 : -1);
}
function translateDiagnostic(input, doc, offset) {
    const start = mapPos(input.line, input.column, doc, offset);
    const result = {
        from: start,
        to: input.endLine != null && input.endColumn != 1 ? mapPos(input.endLine, input.endColumn, doc, offset) : start,
        message: input.message,
        source: input.ruleId ? "eslint:" + input.ruleId : "eslint",
        severity: input.severity == 1 ? "warning" : "error"
    };
    if (input.fix) {
        const { range, text } = input.fix;
        const from = range[0] + offset.pos - start;
        const to = range[1] + offset.pos - start;
        result.actions = [{
                name: "fix",
                apply(view, start) {
                    view.dispatch({ changes: { from: start + from, to: start + to, insert: text }, scrollIntoView: true });
                }
            }];
    }
    return result;
}

export { completionPath, esLint, localCompletionSource, scopeCompletionSource, snippets, spel, spelLanguage };
